#ifdef PRECOMPUTED_NORMALS

Texture2D normalTexture;
Texture2D coarseNormalTexture;

// Computes a blended normal from the current and the coarser levels normal maps.
float3 computeNormal(in VertexIn In)
{
  float3 fineNormal = normalize(normalTexture.Sample(normalSampler, In.TerrainTexCoords.xy).rgb);
  float3 coarseNormal = normalize(coarseNormalTexture.Sample(normalSampler, In.TerrainTexCoords.zw).rgb);

  return normalize(lerp(fineNormal, coarseNormal, In.BlendFactor));
}

#else

float sampleHeight(in Texture2D source, in float2 texCoords, in float2 offset)
{
  return sampleTexture(source, texCoords + offset, float2(0, 0)).r;
}

float sampleHeightNoData(in Texture2D source, in float2 texCoords, in float2 offset, in float defaultValue, out float distance)
{
  float h = sampleLinearNoData(source, coarseHeightSampler, texCoords + offset, noClipmapData);

  bool isValid = h != noClipmapData;
  distance = isValid ? 0 : 1;
  h = isValid ? h : defaultValue;

  return h;
}

float3 computeLevelNormal(in Texture2D source, in float2 texCoords, float2 texelOffset)
{
  float2 dimensions;
  source.GetDimensions(dimensions.x, dimensions.y);

  float2 offsetSize = 1.0 / dimensions * texelOffset;

  float2 dZ;

  if (!enableNoData) {
    float t = sampleHeight(source, texCoords, float2(0.0, offsetSize.y)).r;
    float b = sampleHeight(source, texCoords, float2(0.0, -offsetSize.y)).r;
    float l = sampleHeight(source, texCoords, float2(-offsetSize.x, 0.0)).r;
    float r = sampleHeight(source, texCoords, float2(offsetSize.x, 0.0)).r;

    dZ.x = (l - r) * 0.5;
    dZ.y = (t - b) * 0.5;
  } else {
    float center = sampleHeight(source, texCoords, float2(0.0, 0.0)).r;

    float4 dist = { 0, 0, 0, 0 };
    float l = sampleHeightNoData(source, texCoords, float2(-offsetSize.x, 0), center, dist.x);
    float t = sampleHeightNoData(source, texCoords, float2(0, offsetSize.x), center, dist.y);
    float r = sampleHeightNoData(source, texCoords, float2(offsetSize.x, 0), center, dist.z);
    float b = sampleHeightNoData(source, texCoords, float2(0, -offsetSize.x), center, dist.w);

    dZ.x = (l - r) / ((1 - dist.x) + (1 - dist.z));
    dZ.y = (t - b) / ((1 - dist.y) + (1 - dist.w));
  }

  float2 invScale = float2(heightExaggeration, heightExaggeration) / (levelScaleFactor * levelZeroWorldScaleFactor * worldScale.xy);

  return normalize(float3(dZ.x * invScale.x, dZ.y * invScale.y, 1));
}

float3 computeNormal(in VertexIn In, float normalFactor)
{
  float3 fineNormal = computeLevelNormal(heightTexture, In.TerrainTexCoords.xy, normalFactor);

  if (useBlendRegions && In.BlendFactor != 0) {
    float3 coarseNormal = computeLevelNormal(coarseHeightTexture, In.TerrainTexCoords.zw, fineToCoarseRatio * normalFactor);

    if (In.BlendFactor > 0.999999) {
      // When fully blended, it is assumed that the fine normal is no longer valid, since it lies on
      // the border of the fine clipmap level. Blending with this gives unpredictable results.
      return coarseNormal;
    } else {
      return lerp(fineNormal, coarseNormal, In.BlendFactor);
    }
  } else {
    return fineNormal;
  }
}

#endif
