#ifndef CLIPMAP_SAMPLING_HLSL
#define CLIPMAP_SAMPLING_HLSL

#define CLIPMAP_NODATA_NAN
//#define CLIPMAP_GATHER_BUILTIN;

#ifdef CLIPMAP_GATHER_BUILTIN
float4 gather(in Texture2D source, in SamplerState ss, in float2 texCoords, out float2 f)
{
  float2 dimensions;
  source.GetDimensions(dimensions.x, dimensions.y);

  float2 texelPosition = texCoords * dimensions - 0.5;

  //NOTE: Magic offset apparently needed on nVidia cards. Investigate AMD.
  f = frac(texelPosition + 1 / 512.0);

  return source.Gather(ss, texCoords, 0).wzxy;
}
#else
float4 gather(in Texture2D source, in SamplerState ss, in float2 texCoords, out float2 f)
{
  float2 dimensions;
  source.GetDimensions(dimensions.x, dimensions.y);

  // Find location in whole texel coordinates.
  float2 texelPosition = texCoords * dimensions - 0.5;
  float2 leftTexel = floor(texelPosition);

  // Extract fractional component of texture coordinates used to interpolate between
  // neighboring texels.
  f = frac(texelPosition);

  // Size of a texel in normalized texture coordinates.
  float2 texelSize = 1 / dimensions;

  // Sample origin in normlized texture coordinates. Moved to center of texel to ensure no precision
  // errors.
  float2 sampleOrigin = (leftTexel + 0.5) * texelSize;

  float s1 = source.SampleLevel(ss, sampleOrigin + float2(0, 0), 0).r;
  float s2 = source.SampleLevel(ss, sampleOrigin + float2(texelSize.x, 0), 0).r;
  float s3 = source.SampleLevel(ss, sampleOrigin + float2(0, texelSize.y), 0).r;
  float s4 = source.SampleLevel(ss, sampleOrigin + float2(texelSize.x, texelSize.y), 0).r;

  return float4(s1, s2, s3, s4);
}
#endif

// Custom bilinear sampling function with support for nodata handling.
//
// Avoids the problem of retrieving interpolated texel values that have been calculated from a mix of valid
// and invalid values.
float sampleLinearNoData(in Texture2D source, in SamplerState ss, in float2 texCoords, in float noData)
{
  float2 f;
  float4 s = gather(source, ss, texCoords, f);

#ifdef CLIPMAP_NODATA_NAN
  bool4 isValid = bool4(s.x != noData && isfinite(s.x), s.y != noData && isfinite(s.y), s.z != noData && isfinite(s.z), s.w != noData && isfinite(s.w));
#else
  bool4 isValid = bool4(s.x != noData, s.y != noData, s.z != noData, s.w != noData);
#endif

  s.x = isfinite(s.x) ? s.x : 0;
  s.y = isfinite(s.y) ? s.y : 0;
  s.z = isfinite(s.z) ? s.z : 0;
  s.w = isfinite(s.w) ? s.w : 0;

  if (all(isValid)) {
    return lerp(lerp(s.x, s.y, f.x), lerp(s.z, s.w, f.x), f.y);
  } else if (any(isValid)) {
    float x1 = all(isValid.xy) ? f.x : (isValid.x ? 0 : 1);
    float x2 = all(isValid.zw) ? f.x : (isValid.z ? 0 : 1);
    float y = any(isValid.xy) && any(isValid.zw) ? f.y : any(isValid.xy) ? 0 : 1;
    return lerp(lerp(s.x, s.y, x1), lerp(s.z, s.w, x2), y);
  } else {
    return noData;
  }
}

// Linear interpolation function with support for nodata handling.
float lerpNoData(in float x, in float y, in float blendFactor, in float noData, out float factor)
{
#ifdef CLIPMAP_NODATA_NAN
  bool2 isValid = bool2(x != noData && isfinite(x), y != noData && isfinite(y));
#else
  bool2 isValid = bool2(x != noData, y != noData);
#endif

  if (all(isValid)) {
    factor = 0;
    return lerp(x, y, blendFactor);
  } else if (any(isValid)) {
    factor = isValid.x ? 0 : 1;
    return isValid.x ? x : y;
  } else {
    factor = 1;
    return noData;
  }
}

#endif
