Texture2D waveDerivatives;
SamplerState waveDerivativesSampler;

Texture2D wavePosition;
SamplerState wavePositionSampler;

void baseTileCoords(out int2 tile, out float2 pos, in float2 world)
{
  float2 t = tileCoordsFromWorld.x * world + tileRemainder.xy;
  tile = tileOffset.xy + floor(t);
  pos = t - floor(t);
}

void detailTileCoords(out int2 tile, out float2 pos, in float2 world)
{
  float2 t = tileCoordsFromWorld.y * world + tileRemainder.zw;
  tile = tileOffset.zw + floor(t);
  pos = t - floor(t);
}

float4 tileSamplePosition(float rnd0, float rnd1, float rnd2, float rnd3, in float2 pos)
{
  float scale = oneOverTileExtentAdjust*(1.4 - (0.8 / 287.0)*rnd3);
  float2 shifted = scale*(pos + (1.3 / 287.0)*float2(rnd0, rnd1));
  float twist = (0.3 / 287.0)*(rnd2-143.0) + swellDirection;
  float2 r = float2(cos(twist), sin(twist));
  return float4(r.x*shifted.x - r.y*shifted.y,
                r.y*shifted.x + r.x*shifted.y,
                r.x,
                r.y);
}

void setupTileBlend(out float4 sample00,
                    out float4 sample01,
                    out float4 sample10,
                    out float4 sample11,
                    in int2 tile,
                    in float2 pos)
{
  int2 tile00 = tile + int2(0, 0);
  int2 tile01 = tile + int2(0, 1);
  int2 tile10 = tile + int2(1, 0);
  int2 tile11 = tile + int2(1, 1);

  float4 r0 = permute(permute(float4(tile00.x, tile01.x, tile10.x, tile11.x)) + float4(tile00.y, tile01.y, tile10.y, tile11.y));
  float4 r1 = permute(r0 + float4(1, 1, 1, 1));
  float4 r2 = permute(r1 + float4(1, 1, 1, 1));
  float4 r3 = permute(r2 + float4(1, 1, 1, 1));

  sample00 = tileSamplePosition(r0.x, r1.x, r2.x, r3.x, pos - int2(0, 0));
  sample01 = tileSamplePosition(r0.y, r1.y, r2.y, r3.y, pos - int2(0, 1));
  sample10 = tileSamplePosition(r0.z, r1.z, r2.z, r3.z, pos - int2(1, 0));
  sample11 = tileSamplePosition(r0.w, r1.w, r2.w, r3.w, pos - int2(1, 1));
}

float2 inverseRotateXY(in float2 v, in float2 cossin)
{
  return float2(+cossin.x*v.x + cossin.y*v.y,
                -cossin.y*v.x + cossin.x*v.y);
}

float3 inverseRotateXY(in float3 v, in float2 cossin)
{
  return float3(+cossin.x*v.x + cossin.y*v.y,
                -cossin.y*v.x + cossin.x*v.y,
                v.z);
}

float3 waveDisplace(in float2 world, in float depth, in float3 seabed, in float lod)
{
  if (lod < lodDisplace.x) {
    return float3(world, 0);
  }
  float lodBlend = saturate(lodDisplace.y*(lod - lodDisplace.x));

  int2 baseTile;
  float2 basePos;
  baseTileCoords(baseTile, basePos, world);

  float4 p00, p01, p10, p11;
  setupTileBlend(p00, p01, p10, p11, baseTile, basePos);

  float3 s00 = wavePosition.SampleLevel(wavePositionSampler, p00.xy, 0).xyz;
  float3 s01 = wavePosition.SampleLevel(wavePositionSampler, p01.xy, 0).xyz;
  float3 s10 = wavePosition.SampleLevel(wavePositionSampler, p10.xy, 0).xyz;
  float3 s11 = wavePosition.SampleLevel(wavePositionSampler, p11.xy, 0).xyz;

  float b00 = (1.0 - basePos.x)*(1.0 - basePos.y);
  float b01 = (1.0 - basePos.x)*basePos.y;
  float b10 = basePos.x*(1.0 - basePos.y);
  float b11 = basePos.x*basePos.y;;


  float3 s = (b00*inverseRotateXY(s00, p00.zw) +
              b01*inverseRotateXY(s01, p01.zw) +
              b10*inverseRotateXY(s10, p10.zw) +
              b11*inverseRotateXY(s11, p11.zw));

  if ((0.0 < shoalingEffectFactor) && (depth < 0.99)) {
    float2 q = normalize(seabed.xy);

    float t = atan2(q.y, q.x) / 6.28318530718;


    float3 n = wavePosition.SampleLevel(wavePositionSampler, float2(depth, t), 0).xyz;

    float3 m = float3(n.x*float2(q.x, q.y) +
                      n.y*float2(-q.y, q.x),
                      n.z);

    s = lerp(m, s, shoalingEffectFactor*sqrt(depth));
  }


  return float3(world, 0) + lodBlend * significantWaveHeightWorld * s;
}

float3 waveNormal(out float foam, in float2 world, in float depth, in float3 seabed, in float lod, in float foamFactor)
{
  if (lod < lodNormal.x) {
    foam = 0.0;
    return float3(0, 0, 1);
  }
  float lodNormalBlend = saturate(lodNormal.y*(lod - lodNormal.x));

  int2 baseTile;
  float2 basePos;
  baseTileCoords(baseTile, basePos, world);

  float4 p00, p01, p10, p11;
  setupTileBlend(p00, p01, p10, p11, baseTile, basePos);

  // Swap tiles such that a tile is always sampled from the same sampler,
  // otherwise the hardware LOD-calculation fails at seams.
  if (baseTile.x & 1) {
    float4 t;
    t = p00; p00 = p10; p10 = t;
    t = p01; p01 = p11; p11 = t;
    basePos.x = 1.f - basePos.x;
  }
  if (baseTile.y & 1) {
    float4 t;
    t = p00; p00 = p01; p01 = t;
    t = p10; p10 = p11; p11 = t;
    basePos.y = 1.f - basePos.y;
  }


  float moo = worldSpaceMeter / significantWaveHeightWorld;
  float bah = -0.99*moo;

  float4 s00 = waveDerivatives.Sample(waveDerivativesSampler, p00.xy);
  float4 s01 = waveDerivatives.Sample(waveDerivativesSampler, p01.xy);
  float4 s10 = waveDerivatives.Sample(waveDerivativesSampler, p10.xy);
  float4 s11 = waveDerivatives.Sample(waveDerivativesSampler, p11.xy);

  float3 n00 = cross(float3(moo + max(bah, s00.x), 0.0, s00.z),
                     float3(0.0, moo + max(bah, s00.y), s00.w));

  float3 n01 = cross(float3(moo + max(bah, s01.x), 0.0, s01.z),
                     float3(0.0, moo + max(bah, s01.y), s01.w));

  float3 n10 = cross(float3(moo + max(bah, s10.x), 0.0, s10.z),
                     float3(0.0, moo + max(bah, s10.y), s10.w));

  float3 n11 = cross(float3(moo + max(bah, s11.x), 0.0, s11.z),
                     float3(0.0, moo + max(bah, s11.y), s11.w));

  float b00 = (1.0 - basePos.x)*(1.0 - basePos.y);
  float b01 = (1.0 - basePos.x)*basePos.y;
  float b10 = basePos.x*(1.0 - basePos.y);
  float b11 = basePos.x*basePos.y;;

  float3 s =
    inverseRotateXY(n00, p00.zw)*b00 +
    inverseRotateXY(n01, p01.zw)*b01 +
    inverseRotateXY(n10, p10.zw)*b10 +
    inverseRotateXY(n11, p11.zw)*b11;

  if ((0.0 < shoalingEffectFactor) && (depth < 0.99)) {
    float2 q = normalize(seabed.xy);

    float t = atan2(q.y, q.x) / 6.28318530718;

    float4 r = waveDerivatives.Sample(waveDerivativesSampler, float2(sqrt(depth), t));

    float3 u = float3((moo + max(bah, r.x))*float2(q.x, q.y), r.z);
    float3 v = float3((moo + max(bah, r.y))*float2(-q.y, q.x), r.w);

    float3 n = cross(u, v);

    s = lerp(n, s, shoalingEffectFactor*sqrt(depth));
  }


  float4 detailS00 = float4(0, 0, 0, 0);
  float detailBlend = 0.0;
  if (lodDetail.x < lod) {
    detailBlend = saturate(lodDetail.y*(lod - lodDetail.x));

    int2 detailTile;
    float2 detailPos;
    detailTileCoords(detailTile, detailPos, world);

    detailS00 = waveDerivatives.Sample(waveDerivativesSampler, detailPos);

    float3 n = cross(float3(moo + max(bah, detailS00.x), 0.0, detailS00.z),
                     float3(0.0, moo + max(bah, detailS00.y), detailS00.w));

    s += 0.25*detailBlend*n;
  }

  if ((lodFoam.x < lod) && (0.0 < foamFactor)) {
    float lodFoamBlend = saturate(lodFoam.y*(lod - lodFoam.x));

    // Hackish crest-detection, try to detect where
    // domain shrinks the most.
    float q = (b00 * min(s00.x, s00.y) +
               b01 * min(s01.x, s01.y) +
               b10 * min(s10.x, s10.y) +
               b11 * min(s11.x, s11.y) +
               detailBlend*0.125*min(detailS00.x, detailS00.y));

    foam = lodFoamBlend*(saturate(foamDetectionScale * q - foamErosionValue*(1.0 - foamFactor)));
  }
  else {
    foam = 0.0;
  }

  return normalize(lerp(float3(0, 0, 1), normalize(s), lodNormalBlend));
}
